"""
tint.phase_correlation
======================

Functions for performing phase correlation. Used to predict cell movement
between scans.

"""

from typing import cast, Literal, Optional, overload

import numpy as np
from scipy import ndimage

from .config import ConfigType


def get_ambient_flow(
    obj_extent,
    img1: np.ndarray,
    img2: np.ndarray,
    params: ConfigType,
    grid_size: tuple[float, float, float],
) -> Optional[np.ndarray]:
    """Takes in object extent and two images and returns ambient flow. Margin
    is the additional region around the object used to compute the flow
    vectors."""
    margin_r = params["FLOW_MARGIN"] / grid_size[1]
    margin_c = params["FLOW_MARGIN"] / grid_size[2]
    row_lb = obj_extent["obj_center"][0] - obj_extent["obj_radius"] - margin_r
    row_ub = obj_extent["obj_center"][0] + obj_extent["obj_radius"] + margin_r
    col_lb = obj_extent["obj_center"][1] - obj_extent["obj_radius"] - margin_c
    col_ub = obj_extent["obj_center"][1] + obj_extent["obj_radius"] + margin_c
    row_lb = int(row_lb)
    row_ub = int(row_ub)
    col_lb = int(col_lb)
    col_ub = int(col_ub)

    dims = img1.shape

    row_lb = np.max([row_lb, 0])
    row_ub = np.min([row_ub, dims[0]])
    col_lb = np.max([col_lb, 0])
    col_ub = np.max([col_ub, dims[1]])

    flow_region1 = np.copy(img1[row_lb : row_ub + 1, col_lb : col_ub + 1])
    flow_region2 = np.copy(img2[row_lb : row_ub + 1, col_lb : col_ub + 1])

    flow_region1[flow_region1 != 0] = 1
    flow_region2[flow_region2 != 0] = 1
    return fft_flowvectors(flow_region1, flow_region2)


def fft_flowvectors(
    im1: np.ndarray, im2: np.ndarray, global_shift: bool = False
) -> Optional[np.ndarray]:
    """Estimates flow vectors in two images using cross covariance."""
    if not global_shift and (np.max(im1) == 0 or np.max(im2) == 0):
        return None

    crosscov = fft_crosscov(im1, im2)
    sigma = (1 / 8) * min(crosscov.shape)
    cov_smooth = ndimage.filters.gaussian_filter(crosscov, sigma)
    dims = np.array(im1.shape)
    pshift = np.argwhere(cov_smooth == np.max(cov_smooth))[0]
    pshift = (pshift + 1) - np.round(dims / 2, 0)
    return pshift


def fft_crosscov(im1: np.ndarray, im2: np.ndarray) -> np.ndarray:
    """Computes cross correlation matrix using FFT method."""
    fft1_conj = np.conj(np.fft.fft2(im1))
    fft2 = np.fft.fft2(im2)
    normalize = abs(fft2 * fft1_conj)
    normalize[normalize == 0] = 1  # prevent divide by zero error
    cross_power_spectrum = (fft2 * fft1_conj) / normalize
    crosscov = np.fft.ifft2(cross_power_spectrum)
    crosscov = np.real(crosscov)
    return fft_shift(crosscov)


def fft_shift(fft_mat: np.ndarray) -> np.ndarray:
    """Rearranges the cross correlation matrix so that 'zero' frequency or DC
    component is in the middle of the matrix. Taken from stackoverflow Que.
    30630632."""
    rd2 = int(fft_mat.shape[0] / 2)
    cd2 = int(fft_mat.shape[1] / 2)
    quad1 = fft_mat[:rd2, :cd2]
    quad2 = fft_mat[:rd2, cd2:]
    quad3 = fft_mat[rd2:, cd2:]
    quad4 = fft_mat[rd2:, :cd2]
    centered_t = np.concatenate((quad4, quad1), axis=0)
    centered_b = np.concatenate((quad3, quad2), axis=0)
    centered = np.concatenate((centered_b, centered_t), axis=1)
    return centered


@overload
def get_global_shift(im1: Literal[None], im2: Literal[None]) -> None:
    ...


@overload
def get_global_shift(im1: np.ndarray, im2: Literal[None]) -> None:
    ...


@overload
def get_global_shift(im1: Literal[None], im2: np.ndarray) -> None:
    ...


@overload
def get_global_shift(im1: np.ndarray, im2: np.ndarray) -> np.ndarray:
    ...


def get_global_shift(
    im1: Optional[np.ndarray], im2: Optional[np.ndarray]
) -> Optional[np.ndarray]:
    """Returns standardazied global shift vector. im1 and im2 are full frames
    of raw DBZ values."""
    if im2 is None or im1 is None:
        return None
    shift = fft_flowvectors(im1, im2, global_shift=True)
    return shift



"""
tint.helpers
============

"""
from __future__ import annotations
from datetime import datetime
import string
from typing import Iterator, Optional, Union, cast

import cftime
import numpy as np
import pandas as pd
import xarray as xr
from .grid_utils import parse_grid_datetime, get_grid_size, GridType


class Counter:
    """
    Counter objects generate and keep track of unique cell ids.
    Currently only the uid attribute is used, but this framework can
    accomodate further development of merge/split detection.

    Attributes
    ----------
    uid : int
        Last uid assigned.
    cid : dict
        Record of cell genealogy.

    """

    def __init__(self):
        """uid is an integer that tracks the number of independently formed
        cells. The cid dictionary keeps track of 'children' --i.e., cells that
        have split off from another cell."""
        self.uid: int = -1
        self.cid: dict[str, int] = {}

    def next_uid(self, count: int = 1) -> np.ndarray:
        """Incremented for every new independently formed cell."""
        new_uids = self.uid + np.arange(count) + 1
        self.uid += count
        return np.array([str(uid) for uid in new_uids])

    def next_cid(self, pid: str) -> str:
        """Returns parent uid with appended letter to denote child."""
        if pid in self.cid:
            self.cid[pid] += 1
        else:
            self.cid[pid] = 0
        letter = string.ascii_lowercase[self.cid[pid]]
        return pid + letter


class Record:
    """
    Record objects keep track of information related to the shift correction
    process.

    Attributes
    ----------
    scan : int
        Index of the current scan.
    time : datetime
        Time corresponding to scan.
    interval : timedelta
        Temporal difference between the next scan and the current scan.
    interval_ratio : float
        Ratio of current interval to previous interval.
    grid_size : array of floats
        Length 3 array containing z, y, and x mesh size in meters.
    shifts : dataframe
        Records inputs of shift correction process. See matching.correct_shift.
    new_shfits : dataframe
        Row of new shifts to be added to shifts dataframe.
    correction_tally : dict
        Tallies correction cases for performance analysis.


    Shift Correction Case Guide:
    case0 - new object, local_shift and global_shift disagree, returns global
    case1 - new object, returns local_shift
    case2 - local disagrees with last head and global, returns last head
    case3 - local disagrees with last head, returns local
    case4 - local and last head agree, returns average of both
    case5 - flow regions empty or at edge of frame, returns global_shift

    """

    def __init__(self, grid_obj: GridType) -> None:
        self.scan = -1
        self.time = None
        self.interval = None
        self.interval_ratio = None
        self.grid_size = get_grid_size(grid_obj)
        self.shifts = pd.DataFrame()
        self.new_shifts = pd.DataFrame()
        self.correction_tally = {
            "case0": 0,
            "case1": 0,
            "case2": 0,
            "case3": 0,
            "case4": 0,
            "case5": 0,
        }

    def count_case(self, case_num: int) -> None:
        """Updates correction_tally dictionary. This is used to monitor the
        shift correction process."""
        self.correction_tally["case" + str(case_num)] += 1

    def record_shift(self, corr, gl_shift, l_heads, local_shift, case):
        """Records corrected shift, phase shift, global shift, and last
        heads per object per timestep. This information can be used to
        monitor and refine the shift correction algorithm in the
        correct_shift function."""
        if l_heads is None:
            l_heads = np.ma.array([-999, -999], mask=[True, True])

        new_shift_record = pd.DataFrame()
        new_shift_record["scan"] = [self.scan]
        new_shift_record["uid"] = ["uid"]
        new_shift_record["corrected"] = [corr]
        new_shift_record["global"] = [gl_shift]
        new_shift_record["last_heads"] = [l_heads]
        new_shift_record["phase"] = [local_shift]
        new_shift_record["case"] = [case]

        self.new_shifts = pd.concat([self.new_shifts, new_shift_record])

    def add_uids(self, current_objects):
        """Because of the chronology of the get_tracks process, object uids
        cannot be added to the shift record at the time of correction, so they
        must be added later in the process."""
        if len(self.new_shifts) > 0:
            self.new_shifts["uid"] = current_objects["uid"]
            self.new_shifts.set_index(["scan", "uid"], inplace=True)
            self.shifts = pd.concat([self.shifts, self.new_shifts])
            self.new_shifts = pd.DataFrame()

    def update_scan_and_time(self, grid_obj1, grid_obj2=None):
        """Updates the scan number and associated time. This information is
        used for obtaining object properties as well as for the interval ratio
        correction of last_heads vectors."""
        self.scan += 1
        self.time = parse_grid_datetime(grid_obj1)
        if grid_obj2 is None:
            # tracks for last scan are being written
            return
        time2 = parse_grid_datetime(grid_obj2)
        old_diff = self.interval
        self.interval = time2 - self.time
        try:
            inter_val_s = self.interval.total_seconds()
        except AttributeError:
            inter_val_s = self.interval / np.timedelta64(1, "s")
        if old_diff is not None:
            try:
                diff = old_diff.total_seconds()
            except AttributeError:
                diff = old_diff / np.timedelta64(1, "s")
            self.interval_ratio = inter_val_s / diff


def get_grids(
    group: xr.Dataset,
    slices: tuple[int, int],
    lon: xr.DataArray,
    lat: xr.DataArray,
    times: xr.DataArray,
    varname: str = "rain_rate",
) -> Iterator[GridType]:

    dims = group.variables[varname].dims
    for s in range(slices[0], slices[-1]):
        time = cast(Union[np.datetime64, cftime.datetime], times[s].values)
        try:
            data = group.variables[varname][s].values
        except AttributeError:
            data = group.variables[varname][s]
        if len(data.shape) < 3:
            data = data[np.newaxis, :]
        if isinstance(time, np.datetime64):
            t = datetime.fromtimestamp(time.astype("O") / 1e9)
            time = cftime.DatetimeGregorian(
                t.year, t.month, t.day, t.hour, t.minute, t.second
            )
        yield GridType(
            x=group[dims[-1]],
            y=group[dims[-2]],
            lon=lon,
            lat=lat,
            data=np.ma.masked_invalid(data),
            time=time,
        )



"""
tint.grid_utils
===============

Tools for pulling data from reading data.


"""

from __future__ import annotations
from typing import NamedTuple, Union

import cftime
import numpy as np
import pandas as pd
from scipy import ndimage
import xarray as xr

from .config import ConfigType

GridType = NamedTuple(
    "GridType",
    [
        ("x", xr.DataArray),
        ("y", xr.DataArray),
        ("lon", xr.DataArray),
        ("lat", xr.DataArray),
        ("time", xr.DataArray),
        ("data", cftime.datetime),
    ],
)


def parse_grid_datetime(grid_obj: GridType) -> xr.DataArray:
    """Obtains datetime object from the data dictionary."""
    return grid_obj.time


def get_grid_size(grid_obj: GridType) -> np.ndarray:
    """Calculates grid size per dimension given a grid object."""
    return np.array(grid_obj.data.shape)


def get_radar_info(radar: tuple[float, float]) -> dict[str, float]:
    """Return the longitude/latitude of the radar."""
    info = {"radar_lon": radar[0], "radar_lat": radar[1]}
    return info


def get_grid_alt(
    grid_size: tuple[float, float], alt_meters: Union[float, int] = 1500.0
) -> int:
    """Returns z-index closest to alt_meters."""
    return int(np.round(alt_meters / grid_size[0]))


def get_vert_projection(grid: np.ndarray, thresh: Union[float, int] = 40) -> np.ndarray:
    """Returns boolean vertical projection from grid."""
    return np.any(grid > thresh, axis=0)


def get_filtered_frame(
    grid: np.ndarray, min_size: float, thresh: Union[int, float]
) -> np.ndarray:
    """Returns a labeled frame from gridded radar data. Smaller objects
    are removed and the rest are labeled."""
    echo_height = get_vert_projection(grid, thresh)
    labeled_echo = ndimage.label(echo_height)[0]
    frame = clear_small_echoes(labeled_echo, min_size)
    return frame


def clear_small_echoes(label_image: np.ndarray, min_size: float) -> np.ndarray:
    """Takes in binary image and clears objects less than min_size."""
    flat_image = pd.Series(label_image.flatten())
    flat_image = flat_image[flat_image > 0]
    size_table = flat_image.value_counts(sort=False)
    small_objects = size_table.keys()[size_table < min_size]

    for obj in small_objects:
        label_image[label_image == obj] = 0
    label_image = ndimage.label(label_image)
    return label_image[0]


def extract_grid_data(
    grid_obj: GridType, params: ConfigType
) -> tuple[np.ndarray, np.ndarray]:
    """Return filtered grid frame and raw grid slice at global shift altitude."""
    try:
        masked = grid_obj.data.filled(0)
    except AttributeError:
        masked = grid_obj.data
    raw = masked[0, :, :]
    frame = get_filtered_frame(masked, params["MIN_SIZE"], params["FIELD_THRESH"])
    return raw, frame



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
TintX main tracking utility.

TintX offers a simple ``RunDirectory`` class that uses the ``xarray`` library
to read datasets such as model output data or observational datasets.

The cell tracking as well as visualisation of cell tracks is realised through
instances of the ``RunDirectory`` class.
"""

from .reader import RunDirectory
from tintx import config

__all__ = ["RunDirectory"]

__version__ = "2022.8.2"



from __future__ import annotations
from datetime import datetime
import os
from pathlib import Path
from typing import cast, Any, Optional, Union, Iterator

import pandas as pd
from tqdm.auto import tqdm
import xarray as xr

from .helpers import get_grids, GridType
from .tracks import Cell_tracks
from .visualization import full_domain, plot_traj
from .config import set as set_config


__all__ = ["RunDirectory"]


class RunDirectory(Cell_tracks):
    """
    Create an :class:`RunDirecotry` object from a given xarray dataset.

    The :class:`RunDirectory` object gathers all necessary information on the
    data that is stored in the run directory. Once loaded the most
    important meta data will be stored in the run directory for faster
    access the second time.

    Parameters
    ----------
    var_name: str
        Name of the variable that is tracked
    dataset: xarray.DataArray, xarray.Dataset
        Input dataset/data array
    x_coord: str, default: lon
        Name of the X coordinate array/vector
    y_coord: st, default: lat
        Name of the Y coordinate array/vector
    time_coord: str, default: time
        Name of the time variable.

    Example
    -------

    ::

        import xarray
        from tintx import RunDirectory
        dset = xarray.open_mfdataset("_static/data/CMORPH*.nc", combine="by_coords")
        run = RunDirectory('precip', dset.isel(time=slice(10, 150)))

    """

    @classmethod
    def from_files(
        cls,
        input_files: Union[os.PathLike, list[os.PathLike]],
        var_name: str,
        *,
        x_coord: str = "lon",
        y_coord: str = "lat",
        time_coord: str = "time",
        start: Optional[Union[str, datetime, pd.Timestamp]] = None,
        end: Optional[Union[str, datetime, pd.Timestamp]] = None,
        **kwargs: Any,
    ):
        """
        Create an :class:`RunDirecotry` object from given input file(s)/ directory.

        The :class:`RunDirectory` object gathers all necessary information on the
        data that is stored in the run directory. Once loaded the most
        important meta data will be stored in the run directory for faster
        access the second time.

        Parameters
        ----------
        inp_files: os.PathLike, list[os.PathLike]
            Input filenames or directory that is opened.
        var_name: str
            Name of the variable that is tracked
        start: str, pandas.Timestamp, datetime.datetime (default: None)
            first time step that is considered, if none given the first
            of the dataset is considered.
        end: str, pandas.Timestamp, datetime.datetime (default: None)
            last time step that is considered, if none given the last
            of the dataset is considered.
        x_coord: str (default: lon)
            The name of the longitude vector/array
        x_coord: str (default: lat)
            The name of the latitude vector/array
        time_coord: str, default: time
            The name of the time variable
        kwargs:
            Additional keyword arguments that are passed to open the dataset
            with xarray

        Example
        -------

        ::

            from tintx import RunDirectory
            run = RunDirectory.from_files("_static/data/CMORPH*.nc", "cmorph")

        """
        defaults: dict[str, Union[str, bool]] = dict(
            combine="by_coords",
            use_cftime=True,
        )
        defaults = dict()
        for key, value in defaults.items():
            kwargs.setdefault(key, value)
        _dset = xr.open_mfdataset(input_files, **kwargs)
        start_time = start or _dset[time_coord].isel({time_coord: 0})
        end_time = end or _dset[time_coord].isel({time_coord: -1})
        return cls(
            var_name,
            _dset.sel({time_coord: slice(start_time, end_time)}),
            x_coord=x_coord,
            y_coord=y_coord,
            time_coord=time_coord,
        )

    def __init__(
        self,
        var_name: str,
        dataset: Union[xr.Dataset, xr.DataArray],
        *,
        time_coord: str = "time",
        x_coord: str = "lon",
        y_coord: str = "lat",
    ):
        if isinstance(dataset, xr.DataArray):
            self.data = xr.Dataset({var_name: dataset})
        else:
            self.data = dataset
        self.lons = cast(xr.DataArray, dataset[x_coord])
        self.lats = cast(xr.DataArray, dataset[y_coord])
        self.var_name = var_name
        self.time = cast(xr.DataArray, self.data[time_coord])
        self.start = self.time[0]
        self.end = self.time[-1]
        super().__init__(var_name)

    @property
    def tracks(self) -> pd.DataFrame:
        """A pandas.DataFrame representation of the tracked cells."""
        return self._tracks

    def get_tracks(
        self,
        centre: Optional[tuple[float, float]] = None,
        leave_bar: bool = True,
        **tracking_parameters: float,
    ):
        """Obtains tracks given a list of data arrays. This is the
            primary method of the tracks class. This method makes use of all of the
            functions and helper classes defined above.

        Parameters
        ----------
        centre: tuple, default: None
            The centre of the radar station
        leave_bar: bool, default: True
            Leave the progress bar after tracking is finished
        **tracking_parameters: float
            Overwrite the tint tracking parameters with this values for this
            specific tracking. The defaults will be restored afterwards.
            See :py:mod:`tintx.config` for details


        See also
        --------
        :py:mod:`tintx.config`

        Returns
        -------
        int: Number of unique cells identified by the tracking

        Example
        -------

        ::

            import xarray
            from tintx import RunDirectory
            dset = xarray.open_mfdataset("_static/data/CPOL*.nc",
                                         combine="by_coords")
            run = RunDirectory("radar_estimated_rain_rate",
                               dset,
                               x_coord="longitude",
                               y_coord="latitude")
            num_cells = run.get_tracks(field_thresh=0.01)

        """
        with tqdm(
            self.grids, total=self.time.size - 1, desc="Tracking", leave=leave_bar
        ) as pbar:
            with set_config(**tracking_parameters):
                return self._get_tracks(self.grids, centre, pbar=pbar)

    @property
    def grids(self) -> Iterator[GridType]:
        """Create dictionary holding longitude and latitude information."""
        yield from get_grids(
            self.data,
            (0, self.time.size),
            self.lons,
            self.lats,
            self.time,
            varname=self.var_name,
        )

    def animate(self, **kwargs):
        """
        Create a animation of tracked cells.

        Parameters
        ----------
        altitude : float
            The altitude to be plotted in meters.
        vmin, vmax : float, 0.01
            Minimum values for the colormap.
        vmax : float, 15
            Maximum values for the colormap.
        isolated_only: bool, default: False
            If true, only annotates uids for isolated objects. Only used in 'full'
            style.
        cmap: str, default: Blue
            Colormap used for plotting the tracked fields.
        tracers: bool, default: False
            Plot traces of animated cells
        dt: float, default: 0
            Time shift in hours that is applied to the data, this can be
            useful if data time data in in utc but should be displayed in
            another time zone.
        fps: int, default: 5
            Frames per second for output.
        ax: cartopy.mpl.geoaxes.GeoAxesSubplot, default: None
            Axes object that is used to create the animation. If None (default)
            a new axes object will be created.
        kwargs:
            Additional keyword arguments passed to the basemap plotting routine.

        Returns
        -------
        matplotlib.animation.FuncAnimation:
            See `FuncAnimation <https://matplotlib.org/stable/api/_as_gen/matplotlib.animation.FuncAnimation.html>`_
            for more details.
        """
        return full_domain(self, self.grids, **kwargs)

    def plot_trajectories(self, **kwargs):
        """
        Plot traces of trajectories for each particle.

        This code is a fork of plot_traj method in the plot module from the
        trackpy project see http://soft-matter.github.io/trackpy for more details

        Parameters
        ----------
        label : boolean, default: False
            Set to True to write particle ID numbers next to trajectories.
        cmap : colormap, Default matplotlib.colormap.winter
            Colormap used to color different tracks
        ax: cartopy.mpl.geoaxes.GeoAxesSubplot, default: None
            Axes object that is used to create the animation. If None (default)
            a new axes object will be created.
        uids : list[str], default: None
            a preset of stroms to be drawn, instead of all (default)
        color : str, default: None
            A pre-defined color, if None (default) each track will be assigned a
            different color.
        thresh : float, default: -1
            Plot only trajectories with average intensity above this value.
        mintrace : int, default 2
            Minimum length of a trace to be plotted
        **plot_style :
            Additional keyword arguments passed through to the ``Axes.plot(...)``

        Returns
        -------
        cartopy.mpl.geoaxes.GeoAxesSubplot
        """
        return plot_traj(self.tracks, self.lons, self.lats, **kwargs)
